#
#    CassaFS
#    Copyright (C) 2012  Paul Dwerryhouse <paul@dwerryhouse.com.au>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pycassa.pool import ConnectionPool
from pycassa.columnfamily import ColumnFamily
from pycassa.cassandra.ttypes import NotFoundException
from pycassa.types import *
from time import time
from collections import OrderedDict
import uuid, sys
import os.path

from cassafs import zoolock

rootkey = uuid.UUID('{00000000-0000-0000-0000-000000000000}')

def replace(s,buf,p):
    ns = ""
    n = len(s)
    bn = len(buf)
    if n <= p:
        ns = s + ('\0' * (p-n)) + buf
    else:
        ns = s[0:p] + buf + s[p+bn:]
    return ns

class Cass():
    def __init__(self,config):

        self.KEYSPACE=config.get("main","keyspace")
        self.SERVERS=config.get("main","servers").split(",")
        
        self.POOL = ConnectionPool(self.KEYSPACE, server_list=self.SERVERS)
        
        self.DIRECTORY = ColumnFamily(self.POOL, 'Directory')
        self.INODE = ColumnFamily(self.POOL, 'Inode')
        self.BLOCK = ColumnFamily(self.POOL, 'Block')
        
        self.INODE.column_validators['nlink'] = IntegerType()
        self.INODE.column_validators['mode'] = IntegerType()
        self.INODE.column_validators['rdev'] = IntegerType()
        self.INODE.column_validators['uid'] = IntegerType()
        self.INODE.column_validators['gid'] = IntegerType()
        self.INODE.column_validators['size'] = IntegerType()
        self.INODE.column_validators['blksize'] = IntegerType()
        self.INODE.column_validators['blocks'] = IntegerType()
        self.INODE.column_validators['mtime'] = IntegerType()
        self.INODE.column_validators['ctime'] = IntegerType()
        self.INODE.column_validators['atime'] = IntegerType()
        
        self.INODE.key_validation_class = LexicalUUIDType()
        self.BLOCK.key_validation_class = LexicalUUIDType()
        self.DIRECTORY.key_validation_class = LexicalUUIDType()
        
        self.DIRECTORY.default_validation_class = LexicalUUIDType()
        
        self.BLOCK.column_name_class = IntegerType()
            
    def get_key(self,path):
        key = None
        for p in path.split('/'):
            if p == '':
                key = rootkey
            else:
                try:
                    key = self.DIRECTORY.get(key, columns=[p])[p]
                except KeyError:
                    raise NotFoundException
        return key
    
    def writeblock(self,key, buf, blkid, pos, zh):
        
        zl = zoolock.ZooBlockLock(zh, key, blkid)
        lockfile = zl.get_lock()

        try:
            block = self.get_blocks(key, blkid, blkid)
        except NotFoundException:
            block = OrderedDict([(blkid, b'')])
        
        newblock = replace(block[blkid],buf,pos)
        self.set_block(key, blkid, newblock)
    
        zl.drop_lock(lockfile)

        return len(buf)

    def get_inode(self,key):
        try:
            inode = self.INODE.get(key)
        except NotFoundException:
            return None
    
        return inode
    
    def set_inode(self,key, inode):
        self.INODE.insert(key, inode)
    
    def remove_inode(self,key):
        self.INODE.remove(key)
    
    def get_blocks(self,key, start, finish):
        return self.BLOCK.get(key, column_start=start, column_finish=finish)
    
    def remove_block(self,key):
        self.BLOCK.remove(key)
    
    def remove_blocks(self,key, start, end):
        self.BLOCK.remove(key, range(start,end))
    
    def set_block(self,key, blkid, newblock):
        self.BLOCK.insert(key, { blkid : newblock })
    
    def insert_dir(self,key, direntry):
        self.DIRECTORY.insert(key, direntry)
    
    def get_all_dirs(self,key):
        return self.DIRECTORY.get(key,column_count=2147483647)
    
    def get_dir(self,key, filename):
        return self.DIRECTORY.get(key, columns=[filename])[filename]
    
    def remove_dir(self,key, filename):
        return self.DIRECTORY.remove(key, { filename })
    
    def remove_direntry(self,key):
        return self.DIRECTORY.remove(key)
    
    def utimens(self,path, atime, mtime):
        key = self.get_key(path)
        inode = {}
        inode['atime'] = atime.tv_sec
        inode['mtime'] = mtime.tv_sec
        inode = self.INODE.insert(key,inode)
    
    def create(self,dirname,basename,inode):
        key = uuid.uuid4()
        parentkey = self.get_key(dirname)
        self.INODE.insert(key,inode)
        self.DIRECTORY.insert(parentkey, {basename : key})
        return 0
