#!/usr/bin/env python
#
#    CassaFS
#    Copyright (C) 2012  Paul Dwerryhouse <paul@dwerryhouse.com.au>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import zookeeper
from os.path import basename, dirname
import time
import os
from threading import Condition

ZOO_OPEN_ACL_UNSAFE = {"perms":0x1f, "scheme":"world", "id" :"anyone"}

class ZooLock(object):
    def __init__(self, zh, path, mountpoint="/lock"):
        self.cv = Condition()
        self.zh = zh
        self.path = mountpoint + path
        npath = ''
        for i in self.path.split('/')[1:]:
            npath += '/' + i
            try:
                zookeeper.create(self.zh, npath, '', [ZOO_OPEN_ACL_UNSAFE], 0)
            except zookeeper.NodeExistsException:
                pass

    @staticmethod
    def connect(servers="127.0.0.1:2181"):
        return zookeeper.init(servers)

    def __del__(self):
        path = self.path
        while path != '/':
            try:
                zookeeper.delete(self.zh, path)
            except zookeeper.NotEmptyException:
                break
            path = dirname(path)
            
    def notify(self, zh, unknown1, unknown2, lockfile):
        self.cv.acquire()
        self.cv.notify()
        self.cv.release()

    def get_lock(self):
        lockfile = zookeeper.create(self.zh,self.path + '/guid-lock-','lock', [ZOO_OPEN_ACL_UNSAFE], zookeeper.EPHEMERAL | zookeeper.SEQUENCE)

        while(True):
            children = zookeeper.get_children(self.zh, self.path)
    
            # XXX this can be done more efficiently.
            if children != None:
                children.sort()
                children.reverse()

            found = 0
            for child in children:
                if child < basename(lockfile):
                    found = 1
                    break
    
            if not found:
                return lockfile 
    
            self.cv.acquire()
            if zookeeper.exists(self.zh, self.path + '/' + child, self.notify):

                # Process should wait here until notify() wakes it
                self.cv.wait()
            self.cv.release()

    def drop_lock(self,lockfile):
        zookeeper.delete(self.zh,lockfile)

class ZooPathLock(ZooLock):
    def __init__(self, zh, path, mountpoint="/lock"):
        mnt = "/file" + mountpoint
        super(ZooPathLock, self).__init__(zh, path, mountpoint=mnt)

class ZooInodeLock(ZooLock):
    def __init__(self, zh, inode, mountpoint="/lock"):
        mnt = "/inode" + mountpoint
        path = "/" + str(inode)
        super(ZooInodeLock, self).__init__(zh, path, mountpoint=mnt)

class ZooBlockLock(ZooLock):
    def __init__(self, zh, inode, blkid, mountpoint="/lock"):
        mnt = "/block" + mountpoint
        path = "/" + str(inode) + "/" + str(blkid)
        super(ZooBlockLock, self).__init__(zh, path, mountpoint=mnt)

#zoolock = ZooPathLock(path="/one/two/three/four")
#lockfile = zoolock.get_lock()
#time.sleep(5)
#zoolock.drop_lock(lockfile)
